/*
* Funambol is a mobile platform developed by Funambol, Inc.
* Copyright (C) 2003 - 2007 Funambol, Inc.
*
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU Affero General Public License version 3 as published by
* the Free Software Foundation with the addition of the following permission
* added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
* WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
* WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
*
* This program is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
* details.
*
* You should have received a copy of the GNU Affero General Public License
* along with this program; if not, see http://www.gnu.org/licenses or write to
* the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
* MA 02110-1301 USA.
*
* You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
* 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
*
* The interactive user interfaces in modified source and object code versions
* of this program must display Appropriate Legal Notices, as required under
* Section 5 of the GNU Affero General Public License version 3.
*
* In accordance with Section 7(b) of the GNU Affero General Public License
* version 3, these Appropriate Legal Notices must retain the display of the
* "Powered by Funambol" logo. If the display of the logo is not reasonably
* feasible for technical reasons, the Appropriate Legal Notices must display
* the words "Powered by Funambol".
*/

#ifndef INCL_MESSAGING_MANAGER
#define INCL_MESSAGING_MANAGER
/** @cond DEV */

#include "base/fscapi.h"
#include "base/util/utils.h"
#include "base/util/StringBuffer.h"
#include "mail/MailAccount.h"
#include "spds/FolderData.h"
#include "base/globalsdef.h"
#include "mail/MailAccountManager.h"
#include "pim/mail/AccountProvisioner.h"
#include "pim/mail/MailClientData.h"
#include "pim/mail/MailClient2Server.h"
#include "pim/mail/MailServer2Client.h"

#include "mail/MailSyncSourceConfig.h"
#include "client/MailSourceManagementNode.h"

#include <mapidefs.h>
#include <mapix.h>
#include <string>

using namespace std;
BEGIN_NAMESPACE

class MailServer2Client;
class MailClient2Server;
class ItemData;


/**
* Definition of constants to be used to retrieve the id of the folders/account
*/
#define ROLE_ACCOUNT    "account"
#define ROLE_INBOX      "inbox"
#define ROLE_OUTBOX     "outbox"
#define ROLE_DRAFT      "draft"
#define ROLE_SENT       "sent"
#define ROLE_TRASH      "trash"
#define ROLE_DELETED    "deleted"


/**
 * This class handles all the operation on WinMobile messaging 
 * application, like create/update/delete email accounts, folders and emails.
 * To manage email accounts the accountProvisioner is used.
 */
class __declspec(dllexport) MessagingManager : public MailAccountManager {

public:

     MessagingManager(MailSyncSourceConfig& ssc);
    ~MessagingManager();

	__declspec(dllexport) static MessagingManager* getInstance(MailSyncSourceConfig& ssc);


    /**
     * Logs on to the messaging store
     * Returns 0 if no error
     */
    int logon();

    /** 
     * Logs off the messaging store
     * The pSession pointer is set to NULL.
     */
    void logoff();

	
    // --------------------- EMAIL ACCOUNTS ---------------------
    /**
    * Creates the email account.
    * The new account ID will be set inside the 'account' object.
    * Will use the accountProvisioner object.
    * @param account the account settings informations
    * @return 0 if no errors
    */
    int createClientAccount(MailAccount& account);

    /**
    * Updates an email account.
    * Will use the accountProvisioner object.
    * @param account the account settings informations
    * @return 0 if no errors, -1 if account not found
    */
    int updateClientAccount(const MailAccount& account);

    /**
    * Deletes an email account.
    * Will use the accountProvisioner object.
    * @param accountID the account ID to be removed
    * @return 0 if no errors, -1 if account not found
    */
    int deleteClientAccount(const WCHAR* accountID);

        /**
    * Deletes all email accounts.
    * Will use the accountProvisioner object.
    *
    * @return 0 if no errors, -1 if account not found
    */
    int deleteAllAccount();


    // --------------------- EMAIL FOLDERS ---------------------
    /**
    * Called to create an email folder on the Client.
    * Default folders are already created when parent account is created, 
    * so we just check if this folder's role is one of the default ones.
    * Note: the folder ID is set inside the 'folder' object.
    *
    * @param  folder the FolderData settings informations
    * @return 0 if no errors (the folder's role is one of the default ones)
    */
    int createClientFolder(FolderData& folder);

    /**
    * Called to update an email folder on the Client.
    * Default folders are not changeable, so we just check 
    * if this folder's role is one of the default ones.
    *
    * @param folder the FolderData settings informations
    * @return 0 if no errors
    */
    int updateClientFolder(const FolderData& folder);

    /**
    * Called to delete an email folder on the Client.
    * We don't effectively delete the folder: it will be done when the
    * parent account is deleted. So we just check 
    * if this folder exists in the messaging application.
    *
    * @param folderID  the key of folder to delete
    * @return 0 if no errors, -1 if folder not found
    */
    int deleteClientFolder(const WCHAR* folderID);


    // --------------------- EMAIL ITEMS ---------------------
    /**
     * Adds a new Email item into the parent folder.
     * Note: the new Email ID is set inside the MailClientData object.
     * @return 0 if no errors
     */
    int addEmail(MailClientData* m);      // const StringBuffer& parentFolder ?

    /**
     * Adds a new Email item into the parent folder considering that the attachements
     * are added in the mapi storage (not using the content provider).
     * Note: the new Email ID is set inside the MailClientData object.
     * @return 0 if no errors
     */
    int addEmailWithAttachment(MailClientData* m);    

    /**
     * Updates an Email item into the parent folder.
     * @return 0 if no errors, -1 if Email not found
     */
    int updateEmail(const MailClientData* m);

    /**
     * Gets the MAPI type of a specified object 
     * @param key the ID of the object
     * @return object type (as a MAPI define) 
     */
    unsigned long getObjectType(const wchar_t* objectId);

    /**
     * Deletes either an account/folder/message given its ID.
     * Calls OpenObject to detect the type of the object to remove.
     * If it's an account, will use the accountProvisioner to remove it.
     * @param key the unique object ID to remove
     * @return 0 if no errors
     */
    int deleteObject(const wchar_t* key);

    /**
    * Returns the key of a specific mail folder.
    * @param parentKey   the key of the parent folder
    * @param folderRole  the role of the folder (for instance ROLE_INBOX)
    * @return            the key (b64) of the folder if found, empty string if not found
    */
    wstring getIdOfFolder(const StringBuffer& parentKey, const StringBuffer& folderRole);

    /**
    * Returns the role of a specific mail folder.
    * TODO: FIND A WAY TO GET THE FOLDER'S TYPE DIRECTLY!!!
    *
    * @param folderID  the ID (key in b64) of the folder to query
    * @return          the role of the folder (for instance ROLE_INBOX), empty string if not found
    */
    StringBuffer getFolderRole(const wstring& folderID);


    /**
     * Prepares the mail in the Outbox folder with the latest log attached.
     * @param content the content of the log to send (see ClientSettings::print())
     * @return 0 if no errors
    */
    int sendLog(const StringBuffer& content);

    /**
     * Used for backward compatibility: returns the folder's role (for instance ROLE_INBOX) 
     * from the corresponding letter used in old emails < v8 (for instance "I").
     */
    StringBuffer getFolderRoleFromLetter(const wstring& folderLetter);
    
    /**
     * Used for backward compatibility: returns the account ID (base64 key) for the default
     * account (the one named 'funambol').
     */
    StringBuffer getDefaultAccountID();

    /**
     * Used for backward compatibility: creates the default email account, returns its ID.
     * Will check if account already exists.
     */
    StringBuffer createDefaultAccount();


    /**
     * Get the nth account available
     */
    StringBuffer getAccountByIndex(unsigned ind);

    /**
     * Opens the folder 'folderID' and reads all email messages inside it.
     * Populates the arrayList 'mailItems' with the IDs of all messages found.
     * @param mailItems  [IN-OUT] ArrayList of ItemData to fill
     * @param folderID   the folder's ID (b64 key) to look into
     * @param folderRole the folder's role (ROLE_INBOX, ROLE_OUTBOX,...)
     * @return           0 if no errors
     */
    int getAllMessages(ArrayList& mailItems, const wstring& folderID, const StringBuffer& folderRole);

    /**
     * Fills and returns a (new allocated) MailClientData object, given a ItemData.
     * Proxy method: calls mailClient2Server::completeObject().
     */
    MailClientData* completeObject(const ItemData& item, const bool isUpdate);

    /**
     * Moves the message from outbox folder to Sent folder of the same account.
     * If mail not found in outbox folder, nothing is done (return code = 1).
     * Proxy method: calls mailClient2Server::moveMessageFromOutboxToSent().
     */
    int moveMessageFromOutboxToSent(const WCHAR* messageID);

    /**
     * Returns the parent folder's ID (b64 key), empty string if error.
     * @param messageID  ID of the message
     * @return  the ID of parent folder
     */
    wstring getParentFolderID(const WCHAR* messageID);

    /**
     * The same as createKeyFromIdA but returns a wstring
     */
    static wstring createKeyFromId(const SBinary sbEntry);

    /**
     * Decodes the b64 encoded Id into a SBinary struct. If the decoding
     * fails, the SBinary has sb.cb = 0 and sb.lpb = NULL;
     * The SBinary must be freed by the caller.
     * 
     * @param entryId  the b64 encoded key
     * @return the struct filled with the real binary value to be used for the
     *         message (store, folder, mail) item
     */
    static SBinary createIdFromKey(const wchar_t* entryId);

    /**
     * utiliy methods to check if an account exist on device without
     * referencing to client's config
     */
    void checkMailAccounts();
    bool checkAccount(const wchar_t* accountId);

private:

	static MessagingManager* pinstance;
	
    /**
     * The account provisioner object, used to manage the creation of accounts 
     * hiding the wap provisioning doc to this class. 
     */
    AccountProvisioner accountProvisioner;

    /**
     * The mailServer2Client is used to manage operation on incoming email messages
     * (parse data, set into MailClientData object).
     */
    MailServer2Client mailServer2Client;

    /**
     * The mailClient2Server is used to manage operation on outgoing email messages
     * (format data, read from MailClientData object).
     */
    MailClient2Server mailClient2Server;
	
    /** 
	 * The main MAPI session handle (not connected if NULL)
	 */
    IMAPISession* pSession;
	

    /// Returns (new allocated!) IMsgStore from the b64 key
    IMsgStore* getIMsgStore(const WCHAR* key);

    /// Returns  (new allocated!) IMAPIFolder from the b64 key
    IMAPIFolder* getIMapiFolder(const WCHAR* key);

    /// Returns (new allocated!) IMessage from the b64 key
    IMessage* getIMessage(const wchar_t* key);


    /**
    * Opens a generic object given the id (in b64 from server) of that object.
    * The main treated method is the IMAPISessiong->OpenStore.
    * This function is used by wrapper as getIMsgStore, getIFolder, getIMessage.
    * @param sb            ID of the object requested
    * @param requestedType [OUT] the object type returned, as defined in mapiDefs.h (i.e. MAPI_FOLDER)
    */
    LPUNKNOWN* openObject(SBinary sb, unsigned long* requestedType);

    /**
    * Open object usign the key in b64
    * @param key            key in b64, the ID of the object requested
    * @param requestedType [OUT] the object type returned, as defined in mapiDefs.h (i.e. MAPI_FOLDER)
    */
    LPUNKNOWN* openObject(const wchar_t* key, unsigned long* requestedType);

    /**
    * Delete the passed email message given its id (SBinary format)..
    * It uses a particular property of the objects named PR_PARENT_ENTRYID that 
    * provides the id of its parent object, so there's no need to specify the parent.
    * @param pMessage  the IMessage to remove
    * #param   mailId  the message ID in binary format
    * @return  0 if no errors
    */
    int deleteMessage(IMessage* pMessage, SBinary mailId);


    /**
    * Encodes the SBinary.lpb binary value into b64 to be treated
    * in more redeable way (to be echanged as id in syncml process).
    *
    * @param  the SBinary struct retrieved by the store, folder, mail item
    * @return the encoded b64 value of the binary id. Return StringBuffer empty if SBinary.cb is 0
    */ 
    static StringBuffer createKeyFromIdA(const SBinary sbEntry);


    /**
    * Gets the ID of the create account given the GUID. It is the id that must
    * be exchanged with the server.
    * @param guid the account GUID
    * @return     the account unique key (ID)
    */ 
    wstring getAccountID(const wchar_t* guid);

    /**
    * This function return a wstring containing the MS GUID of the account given 
    * its unique ID (key) which is the one exchanged with the server.
    * The GUID is to used to modify/delete the account using the provisioning method.
    * @param key  the account unique key (ID)
    * @return     the account GUID
    */
    wstring getAccountGUID(const wchar_t* key);
    
    /**
     * Given an account name return its GUID
     *
     * @param accountName   the account name (MAPI display name)
     * @return              the account GUID
     */
    //wstring getAccountGuidByName(const wchar_t* accountName);
    bool getAccountGuidByName(const wchar_t* accountName, wstring* guid); 
     /**
     * Returns the folder tag entry_id given the folder's role.
     * @param t  the folder's role
     * @return   the folder tag entry_id
     */
    ULONG getFolderEntryId(const StringBuffer& t);

	/**
	 * deletes a just created email that cannot be completed
	 * due to low memory on the device or other errors 
     */
	HRESULT deleteStoredMessage(LPMAPIFOLDER folder, SPropValue* rgprops);

    /**
     * Returns the parent folder of the passed message, NULL if error.
     * Note: the returned IMAPIFolder* must be released by the caller.
     */
    IMAPIFolder* getParentFolder(IMessage* pMessage);


    /**
    * Returns the key of a specific mail folder.
    * @param folder   the IMAPIFolder* object of the folder to query
    * @return         the key (b64) of the folder if found, empty string if error
    */
    wstring getEntryIdOfFolder(IMAPIFolder* folder);

    /**
     * Moves a message to a specific mail folder.
     * @param messageID     message ID
     * @param sourceFolder  source IMAPIFolder* object 
     * @param destFolder    destination IMAPIFolder* object
     * @return              S_OK for success otherwise S_FALSE        
     */
    HRESULT moveMessage(const WCHAR* messageID, IMAPIFolder* sourceFolder, IMAPIFolder* destFolder);

    /**
     * Add an email on native mapi store. 
     * @param m             the object containing the value to be set in the email
     * @param addAttachemnt true means that if there are attachments in the values of the MailClientData
     *                      they are added in the store. Basically addAttachments=true is used in the 
     *                      sendLog function to add properly the attachment in the Outbox folder.
     *                      Usually the addAttachement=false is the default
     * @return              0 if no errors
     */
    int addEmailOnMapiStore(MailClientData* m, bool addAttachments);

};

/**
 * exported singleton methods
 */
__declspec(dllexport) MessagingManager* getMessagingManager(/*MailSyncSourceConfig& ssc*/);

END_NAMESPACE
#endif
