/*
* Funambol is a mobile platform developed by Funambol, Inc.
* Copyright (C) 2009 Funambol, Inc.
*
* This program is free software; you can redistribute it and/or modify it under
* the terms of the GNU Affero General Public License version 3 as published by
* the Free Software Foundation with the addition of the following permission
* added to Section 15 as permitted in Section 7(a): FOR ANY PART OF THE COVERED
* WORK IN WHICH THE COPYRIGHT IS OWNED BY FUNAMBOL, FUNAMBOL DISCLAIMS THE
* WARRANTY OF NON INFRINGEMENT  OF THIRD PARTY RIGHTS.
*
* This program is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
* FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
* details.
*
* You should have received a copy of the GNU Affero General Public License
* along with this program; if not, see http://www.gnu.org/licenses or write to
* the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
* MA 02110-1301 USA.
*
* You can contact Funambol, Inc. headquarters at 643 Bair Island Road, Suite
* 305, Redwood City, CA 94063, USA, or at email address info@funambol.com.
*
* The interactive user interfaces in modified source and object code versions
* of this program must display Appropriate Legal Notices, as required under
* Section 5 of the GNU Affero General Public License version 3.
*
* In accordance with Section 7(b) of the GNU Affero General Public License
* version 3, these Appropriate Legal Notices must retain the display of the
* "Powered by Funambol" logo. If the display of the logo is not reasonably
* feasible for technical reasons, the Appropriate Legal Notices must display
* the words "Powered by Funambol".
*/

#ifndef _INCL_MESSAGING_STATUS_H_
#define _INCL_MESSAGING_STATUS_H_

#include "base/util/utils.h"
#include "base/util/ArrayElement.h"
#include "base/util/StringBuffer.h"
#include "mail/MailSyncSourceConfig.h"
#include "mail/MailAccount.h"
#include "spds/FolderData.h"
#include "base/globalsdef.h"
#include "mail/MailAccountManager.h"
#include "base/fscapi.h"
#include <windows.h>    // semaphore object
#include <mapidefs.h>
#include <string>

using namespace std;

BEGIN_NAMESPACE

typedef enum mapi_item_status_t {
    item_status_unset = 0,
    item_status_new,
    item_status_updated,
    item_status_deleted
} MapiItemStatus;

#define MESSAGE_STATUS_CACHE_NAME       "message_status.cache"
#define MESSAGE_STATUS_CACHE_TEMP_NAME  "message_status.cache.tmp"

// string identifiers for object serialization into cache
#define MAPI_MESSAGE_ID            "M"
#define MAPI_FOLDER_ID             "F"
#define MAPI_ACCOUNT_ID            "A"

#define STATUS_NEW_ID              "N"
#define STATUS_UPDATED_ID          "U"
#define STATUS_DELETED_ID          "D"

#define PARENT_ROLE_INBOX          "I"
#define PARENT_ROLE_OUTBOX         "O"

class ItemData : public ArrayElement {

private:

    wstring         id;
    wstring         parent_id;

    // object type (supported types: MAPI_STORE, MAPI_FOLDER or MAPI_MESSAGE)
    unsigned long   type;

    // object status: created, updated or deleted
    MapiItemStatus  status;

    StringBuffer    parent_role;
    bool            message_read; // mark message as read
    bool            incomplete;   // mark message as incomplete

public:

    ItemData(const wstring& id_, const wstring& parentId, 
             unsigned long item_type, MapiItemStatus status_, 
             const StringBuffer& parentRole="", bool read_flag = false, bool incomplete_flag = false) : 
             id(id_), parent_id(parentId), 
             type(item_type), status(status_), parent_role(parentRole), 
             message_read(read_flag), incomplete(incomplete_flag) {}
        
    ItemData(const wstring& id_, unsigned long item_type) : id(id_), parent_id(L""), 
        type(item_type), status(item_status_unset), parent_role(""), 
        message_read(false), incomplete(false) {}

    ~ItemData() {}
    void setId(const wstring& _id) { id = _id; }
    void setParentId(const wstring& parentId) { parent_id = parentId; }
    void setStatus(const MapiItemStatus item_status) { status = item_status; }
    void setParentRole(const StringBuffer& parentRole) { parent_role = parentRole; }
    void setReadFlag(bool readFlag) { message_read = readFlag; }
    void setIncompleteFlag(bool incompleteFlag) { incomplete = incompleteFlag; }

    const wchar_t*  getId() const { return id.c_str(); }
    const wchar_t*  getParentId() const { return parent_id.c_str(); }
    unsigned long   getType() const { return type; }
    const StringBuffer& getParentRole() const { return parent_role; }
    MapiItemStatus  getStatus() const { return status; }
    bool            getReadFlag() const { return message_read; }
    bool            getIncompleteFlag() const { return incomplete; }

    ArrayElement* clone() { return new ItemData(*this); }
};


class MessagingStatus {

public:
    /**
     * singleton methods 
     */
    static MessagingStatus* getInstance();
    static void disposeInstance();
    
    /**
     * remove element from cache without persisting
     * it to file system
     */
    bool removeCacheItem(const wstring& id);

    /**
     * force cache store on file system and clean up memory
     */
    bool commit();

    /**
     * clean up memory cache and delete store on filesystem
     */
    bool resetCache();
    
    /**
     * check if cache file exist
     */
    bool checkCacheFile() const;

    /**
     * exported methods
     */
    bool addMessagingStatusItem(const ItemData& item);
    bool updateMessagingStatusItem(const ItemData& item);
    const ArrayList& getMessagingStatusNewItems();
    const ArrayList& getMessagingStatusUpdatedItems();
    const ArrayList& getMessagingStatusDeletedItems();
    const ItemData* getMessagingStatusItem(const wstring& itemId);

    bool lockMessagingStatusCache();
    bool unlockMessagingStatusCache();
    bool refreshCache();
    bool checkMessagingStatusCache() const;

private:
    MessagingStatus();        
    ~MessagingStatus();

    // disable copy construct
    MessagingStatus(const MessagingStatus& msgStatus);
    // disable copy 
    MessagingStatus& operator=(const MessagingStatus& msgStatus);

    const ArrayList& getNewItems();
    const ArrayList& getUpdatedItems();
    const ArrayList& getDeletedItems();

    /**
     * insert element in cache and write down to filesystem
     */
    bool addItem(const ItemData& itemData);

    /**
     * update element in cache (or add if not found) and 
     * persist cache
     */
    bool updateItem(const ItemData& itemData);

    /**
     * populate memory cache with data from file system
     */
    bool readCache();
    
    /**
     * synchronize memory cache with filesystem cache store:
     * if there are no items in memory stored cache is removed
     */
    bool storeCache();

    /**
     * remove cache file
     */
    bool removeCache();

     /**
     *  set semaphore lock for status cache file
     */
    bool lockCache();

    /**
     * unset semaphore lock for status caache file
     */
    bool unlockCache(); 

    // get a cache item by id
    const ItemData* getItem(const wstring& itemId);

    // find items in status cache by ID
    int findItem(const wstring& itemId) const;
    
    // fill up temporary caches 
    void getItemsByStatus(ArrayList& cache, MapiItemStatus status);
    
    // serialize data on file system
    bool formatCacheLine(const ItemData* item, StringBuffer& cacheLine);


    // check status of cache lock
    bool checkCacheLock(unsigned long timeout=0L);

    // full cache name on file system
    StringBuffer cacheName;
    
    // global semaphore for concurrent access to cache file
    HANDLE cacheSemaphore;

    // global semaphore name;
    static const wchar_t* cacheSemaphoreName;
    // lock request semaphore timeout 
    static const unsigned int semaphoreTimeOut;

    // list of mapi items in cache: persisted 
    // on every add/update operation
    ArrayList itemStatusCache;

    // temporary caches 
    ArrayList newItemsCache;
    ArrayList updatedItemsCache;
    ArrayList deletedItemsCache;

    // number of tokens (which make up a complete MAPI item) 
    // contained in a single cache line
    static const unsigned messageStatusCacheItemsPerLine = 7;

    // singleton instance
    static MessagingStatus* pinstance;
};

MessagingStatus* getMessagingStatus();

#endif // _INCL_MESSAGING_STATUS_H_
